from jinja2 import Template

import logging


from servicecatalog_puppet.asset_helpers import read_from_site_packages
from servicecatalog_puppet.constants import BOOTSTRAP_STACK_NAME
from servicecatalog_puppet.constants import PUPPET_ORG_ROLE_FOR_EXPANDS_ARN

logger = logging.getLogger(__file__)


def do_bootstrap_org_master(puppet_account_id, cloudformation, puppet_version):
    logger.info('Starting bootstrap of org master')
    stack_name = "{}-org-master".format(BOOTSTRAP_STACK_NAME)
    template = read_from_site_packages('{}.template.yaml'.format(stack_name))
    template = Template(template).render(VERSION=puppet_version)
    args = {
        'StackName': stack_name,
        'TemplateBody': template,
        'Capabilities': ['CAPABILITY_NAMED_IAM'],
        'Parameters': [
            {
                'ParameterKey': 'PuppetAccountId',
                'ParameterValue': str(puppet_account_id),
            }, {
                'ParameterKey': 'Version',
                'ParameterValue': puppet_version,
                'UsePreviousValue': False,
            },
        ],
    }
    cloudformation.create_or_update(**args)
    response = cloudformation.describe_stacks(StackName=stack_name)
    if len(response.get('Stacks')) != 1:
        raise Exception("Expected there to be only one {} stack".format(stack_name))
    stack = response.get('Stacks')[0]

    for output in stack.get('Outputs'):
        if output.get('OutputKey') == PUPPET_ORG_ROLE_FOR_EXPANDS_ARN:
            logger.info('Finished bootstrap of org-master')
            return output.get("OutputValue")

    raise Exception("Could not find output: {} in stack: {}".format(PUPPET_ORG_ROLE_FOR_EXPANDS_ARN, stack_name))




import os

PREFIX = 'SC-P--'
BOOTSTRAP_STACK_NAME = "servicecatalog-puppet"
SERVICE_CATALOG_PUPPET_REPO_NAME = "ServiceCatalogPuppet"
OUTPUT = "output"
TEMPLATES = os.path.sep.join([OUTPUT, "templates"])
LAUNCHES = os.path.sep.join([OUTPUT, "launches"])
CONFIG_PARAM_NAME = "/servicecatalog-puppet/config"
CONFIG_PARAM_NAME_ORG_IAM_ROLE_ARN = "/servicecatalog-puppet/org-iam-role-arn"
PUPPET_ORG_ROLE_FOR_EXPANDS_ARN = "PuppetOrgRoleForExpandsArn"
HOME_REGION_PARAM_NAME = "/servicecatalog-puppet/home-region"



from copy import deepcopy

from servicecatalog_puppet.macros import macros

import logging

logger = logging.getLogger(__file__)


def expand_path(account, client):
    ou = client.convert_path_to_ou(account.get('ou'))
    account['ou'] = ou
    return expand_ou(account, client)


def expand_ou(original_account, client):
    expanded = []
    response = client.list_children_nested(ParentId=original_account.get('ou'), ChildType='ACCOUNT')
    for result in response:
        new_account_id = result.get('Id')
        response = client.describe_account(AccountId=new_account_id)
        new_account = deepcopy(original_account)
        del new_account['ou']
        if response.get('Account').get('Name') is not None:
            new_account['name'] = response.get('Account').get('Name')
        new_account['email'] = response.get('Account').get('Email')
        new_account['account_id'] = new_account_id
        new_account['expanded_from'] = original_account.get('ou')
        new_account['organization'] = response.get('Account').get('Arn').split(":")[5].split("/")[1]
        expanded.append(new_account)
    return expanded


def do_expand(manifest, client):
    new_manifest = deepcopy(manifest)
    new_accounts = new_manifest['accounts'] = []

    logger.info('Starting the expand')

    for account in manifest.get('accounts'):
        if account.get('account_id'):
            logger.info("Found an account: {}".format(account.get('account_id')))
            new_accounts.append(account)
        elif account.get('ou'):
            ou = account.get('ou')
            logger.info("Found an ou: {}".format(ou))
            if ou.startswith('/'):
                new_accounts += expand_path(account, client)
            else:
                new_accounts += expand_ou(account, client)

    logger.debug(new_accounts)

    for parameter_name, parameter_details in new_manifest.get('parameters', {}).items():
        if parameter_details.get('macro'):
            macro_to_run = macros.get(parameter_details.get('macro').get('method'))
            result = macro_to_run(client, parameter_details.get('macro').get('args'))
            parameter_details['default'] = result
            del parameter_details['macro']

    for first_account in new_accounts:
        for parameter_name, parameter_details in first_account.get('parameters', {}).items():
            if parameter_details.get('macro'):
                macro_to_run = macros.get(parameter_details.get('macro').get('method'))
                result = macro_to_run(client, parameter_details.get('macro').get('args'))
                parameter_details['default'] = result
                del parameter_details['macro']

        times_seen = 0
        for second_account in new_accounts:
            if first_account.get('account_id') == second_account.get('account_id'):
                times_seen += 1
                if times_seen > 1:
                    message = "{} has been seen twice.".format(first_account.get('account_id'))
                    if first_account.get('expanded_from'):
                        message += "  It was included due to it being in the ou: {}".format(
                            first_account.get('expanded_from')
                        )
                    if second_account.get('expanded_from'):
                        message += "  It was included due to it being in the ou: {}".format(
                            second_account.get('expanded_from')
                        )
                    raise Exception(message)

    for launch_name, launch_details in new_manifest.get('launches').items():
        for parameter_name, parameter_details in launch_details.get('parameters', {}).items():
            if parameter_details.get('macro'):
                macro_to_run = macros.get(parameter_details.get('macro').get('method'))
                result = macro_to_run(client, parameter_details.get('macro').get('args'))
                parameter_details['default'] = result
                del parameter_details['macro']

    return new_manifest



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from pytest import fixture


@fixture
def all_tasks():
    return {
        "012345678910-eu-west-1-all-enable-config": {
            "launch_name": "all-enable-config",
            "portfolio": "ccoe-mandatory-product-portfolio",
            "product": "Config-Enable",
            "version": "v1",
            "product_id": "prod-bbbbbbbbbbb",
            "version_id": "pa-bbbbbbbbbbb",
            "account_id": "012345678910",
            "region": "eu-west-1",
            "puppet_account_id": "012345678911",
            "parameters": [],
            "ssm_param_inputs": [],
            "depends_on": [],
            "dependencies": []
        },
        "012345678910-eu-west-1-all-rules-config": {
            "launch_name": "all-rules-config",
            "portfolio": "ccoe-mandatory-product-portfolio",
            "product": "Config-Rules",
            "version": "v1",
            "product_id": "prod-ccccccccccccc",
            "version_id": "pa-ccccccccccccc",
            "account_id": "012345678910",
            "region": "eu-west-1",
            "puppet_account_id": "012345678911",
            "parameters": [],
            "ssm_param_inputs": [],
            "depends_on": [
                "all-enable-config"
            ],
            "dependencies": []
        },
        "012345678912-eu-west-1-all-enable-config": {
            "launch_name": "all-enable-config",
            "portfolio": "ccoe-mandatory-product-portfolio",
            "product": "Config-Enable",
            "version": "v1",
            "product_id": "prod-bbbbbbbbbbb",
            "version_id": "pa-bbbbbbbbbbb",
            "account_id": "012345678912",
            "region": "eu-west-1",
            "puppet_account_id": "012345678911",
            "parameters": [],
            "ssm_param_inputs": [],
            "depends_on": [],
            "dependencies": []
        },
    }


@fixture
def sut():
    from servicecatalog_puppet import cli
    return cli


def test_wire_dependencies(sut, all_tasks):
    # setup
    expected_result = [
        {
            "launch_name": "all-enable-config",
            "portfolio": "ccoe-mandatory-product-portfolio",
            "product": "Config-Enable",
            "version": "v1",
            "product_id": "prod-bbbbbbbbbbb",
            "version_id": "pa-bbbbbbbbbbb",
            "account_id": "012345678910",
            "region": "eu-west-1",
            "puppet_account_id": "012345678911",
            "parameters": [],
            "ssm_param_inputs": [],
            "dependencies": []
        },
        {
            "launch_name": "all-rules-config",
            "portfolio": "ccoe-mandatory-product-portfolio",
            "product": "Config-Rules",
            "version": "v1",
            "product_id": "prod-ccccccccccccc",
            "version_id": "pa-ccccccccccccc",
            "account_id": "012345678910",
            "region": "eu-west-1",
            "puppet_account_id": "012345678911",
            "parameters": [],
            "ssm_param_inputs": [],
            "dependencies": [
                {
                    "launch_name": "all-enable-config",
                    "portfolio": "ccoe-mandatory-product-portfolio",
                    "product": "Config-Enable",
                    "version": "v1",
                    "product_id": "prod-bbbbbbbbbbb",
                    "version_id": "pa-bbbbbbbbbbb",
                    "account_id": "012345678910",
                    "region": "eu-west-1",
                    "puppet_account_id": "012345678911",
                    "parameters": [],
                    "ssm_param_inputs": [],
                    "dependencies": []
                },
                {
                    "launch_name": "all-enable-config",
                    "portfolio": "ccoe-mandatory-product-portfolio",
                    "product": "Config-Enable",
                    "version": "v1",
                    "product_id": "prod-bbbbbbbbbbb",
                    "version_id": "pa-bbbbbbbbbbb",
                    "account_id": "012345678912",
                    "region": "eu-west-1",
                    "puppet_account_id": "012345678911",
                    "parameters": [],
                    "ssm_param_inputs": [],
                    "dependencies": []
                },
            ]
        },
        {
            "launch_name": "all-enable-config",
            "portfolio": "ccoe-mandatory-product-portfolio",
            "product": "Config-Enable",
            "version": "v1",
            "product_id": "prod-bbbbbbbbbbb",
            "version_id": "pa-bbbbbbbbbbb",
            "account_id": "012345678912",
            "region": "eu-west-1",
            "puppet_account_id": "012345678911",
            "parameters": [],
            "ssm_param_inputs": [],
            "dependencies": []
        }
    ]

    # exercise
    actual_result = sut.wire_dependencies(all_tasks)

    # verify
    assert expected_result[0] == actual_result[0]
    e = expected_result[1]
    a = actual_result[1]
    assert e == a
    assert expected_result[2] == actual_result[2]



import yaml
import logging

logger = logging.getLogger(__file__)


def load(f):
    return yaml.safe_load(f.read())


def group_by_tag(launches):
    logger.info('Grouping launches by tag')
    launches_by_tag = {}
    for launch_name, launch_details in launches.items():
        launch_details['launch_name'] = launch_name
        launch_tags = launch_details.get('deploy_to').get('tags', [])
        for tag_detail in launch_tags:
            tag = tag_detail.get('tag')
            if launches_by_tag.get(tag) is None:
                launches_by_tag[tag] = []
            launches_by_tag[tag].append(launch_details)
    logger.info('Finished grouping launches by tag')
    return launches_by_tag


def group_by_account(launches):
    logger.info('Grouping launches by account')
    launches_by_account = {}
    for launch_name, launch_details in launches.items():
        launch_details['launch_name'] = launch_name
        launch_accounts = launch_details.get('deploy_to').get('accounts', [])
        for account_detail in launch_accounts:
            account_id = account_detail.get('account_id')
            if launches_by_account.get(account_id) is None:
                launches_by_account[account_id] = []
            launches_by_account[account_id].append(launch_details)
    logger.info('Finished grouping launches by account')
    return launches_by_account


def group_by_product(launches):
    logger.info('Grouping launches by product')
    launches_by_product = {}
    for launch_name, launch_details in launches.items():
        product = launch_details.get('product')
        if launches_by_product.get(product) is None:
            launches_by_product[product] = []
        launch_details['launch_name'] = launch_name
        launches_by_product[product].append(launch_details)
    logger.info('Finished grouping launches by product')
    return launches_by_product


def generate_launch_map(accounts, launches_by_account, launches_by_tag):
    logger.info('Generating launch map')
    deployment_map = {}
    for account in accounts:
        account_id = account.get('account_id')
        deployment_map[account_id] = account
        launches = account['launches'] = {}
        for launch in launches_by_account.get(account_id, []):
            launch['match'] = "account_match"
            launches[launch.get('launch_name')] = launch
        for tag in account.get('tags'):
            for launch in launches_by_tag.get(tag, []):
                launch['match'] = "tag_match"
                launch['matching_tag'] = tag
                launches[launch.get('launch_name')] = launch
    logger.info('Finished generating launch map')
    return deployment_map


def build_deployment_map(manifest):
    accounts = manifest.get('accounts')
    launches = manifest.get('launches')

    verify_no_ous_in_manifest(accounts)

    launches_by_product = group_by_product(launches)
    # check_for_duplicate_products_in_launches(launches_by_product)
    launches_by_tag = group_by_tag(launches)
    launches_by_account = group_by_account(launches)

    return generate_launch_map(
        accounts,
        launches_by_account,
        launches_by_tag,
    )


def verify_no_ous_in_manifest(accounts):
    for account in accounts:
        if account.get('account_id') is None:
            raise Exception("{} account object does not have an account_id".format(account.get('name')))


def check_for_duplicate_products_in_launches(launches_by_product):
    logger.info('Checking for duplicate products by tag')
    for product_name, product_launches in launches_by_product.items():
        tags_seen = {}
        for product_launch in product_launches:
            for tag in product_launch.get('deploy_to').get('tags', []):
                tag_name = tag.get('tag')
                if tags_seen.get(tag_name) is None:
                    tags_seen[tag_name] = product_launch
                else:
                    raise Exception(
                        "Cannot process {}.  Already added {} because of tag: {}".format(
                            product_launch.get('launch_name'),
                            tags_seen[tag_name].get('launch_name'),
                            tag_name
                        )
                    )
    logger.info('Finished checking for duplicate products by tag')

    logger.info('Checking for duplicate products by account listed twice')
    for product_name, product_launches in launches_by_product.items():
        accounts_seen = {}
        for product_launch in product_launches:
            for account in product_launch.get('deploy_to').get('accounts', []):
                account_id = account.get('account_id')
                if accounts_seen.get(account_id) is None:
                    accounts_seen[account_id] = product_launch
                else:
                    raise Exception(
                        "Cannot process {}. {} is already receiving product: {} as it was listed in launch: {}".format(
                            product_launch.get('launch_name'),
                            account_id,
                            accounts_seen[account_id].get('product'),
                            accounts_seen[account_id].get('launch_name'),
                        )
                    )
    logger.info('Finished checking for duplicate products by account listed twice')



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


with open("servicecatalog_puppet/requirements.txt", "r") as fh:
    requirements = fh.read().split("\n")

setuptools.setup(
    name="aws-service-catalog-puppet",
    version="0.1.5",
    author="Eamonn Faherty",
    author_email="aws-service-catalog-tools@amazon.com",
    description="Making it easier to deploy ServiceCatalog products",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/awslabs/aws-service-catalog-puppet-framework",
    packages=setuptools.find_packages(),
    package_data={'servicecatalog_puppet': ['*','*/*','*/*/*']},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    entry_points={
        'console_scripts': [
            'servicecatalog-puppet = servicecatalog_puppet.cli:cli'
        ]},
    install_requires=requirements,
)



import json
import os

import click
from betterboto import client as betterboto_client
from colorclass import Color
from terminaltables import AsciiTable

from servicecatalog_puppet.constants import LAUNCHES
from servicecatalog_puppet.core import get_regions
from servicecatalog_puppet.utils.manifest import build_deployment_map

import logging

logger = logging.getLogger(__file__)


def do_list_launches(manifest):
    click.echo("Getting details from your account...")
    ALL_REGIONS = get_regions(os.environ.get("AWS_DEFAULT_REGION"))
    deployment_map = build_deployment_map(manifest)
    account_ids = [a.get('account_id') for a in manifest.get('accounts')]
    deployments = {}
    for account_id in account_ids:
        for region_name in ALL_REGIONS:
            role = "arn:aws:iam::{}:role/{}".format(account_id, 'servicecatalog-puppet/PuppetRole')
            logger.info("Looking at region: {} in account: {}".format(region_name, account_id))
            with betterboto_client.CrossAccountClientContextManager(
                    'servicecatalog', role, 'sc-{}-{}'.format(account_id, region_name), region_name=region_name
            ) as spoke_service_catalog:

                response = spoke_service_catalog.list_accepted_portfolio_shares()
                portfolios = response.get('PortfolioDetails', [])

                response = spoke_service_catalog.list_portfolios()
                portfolios += response.get('PortfolioDetails', [])

                for portfolio in portfolios:
                    portfolio_id = portfolio.get('Id')
                    response = spoke_service_catalog.search_products_as_admin(PortfolioId=portfolio_id)
                    for product_view_detail in response.get('ProductViewDetails', []):
                        product_view_summary = product_view_detail.get('ProductViewSummary')
                        product_id = product_view_summary.get('ProductId')
                        response = spoke_service_catalog.search_provisioned_products(
                            Filters={'SearchQuery': ["productId:{}".format(product_id)]})
                        for provisioned_product in response.get('ProvisionedProducts', []):
                            launch_name = provisioned_product.get('Name')
                            status = provisioned_product.get('Status')

                            provisioning_artifact_response = spoke_service_catalog.describe_provisioning_artifact(
                                ProvisioningArtifactId=provisioned_product.get('ProvisioningArtifactId'),
                                ProductId=provisioned_product.get('ProductId'),
                            ).get('ProvisioningArtifactDetail')

                            if deployments.get(account_id) is None:
                                deployments[account_id] = {'account_id': account_id, 'launches': {}}

                            if deployments[account_id]['launches'].get(launch_name) is None:
                                deployments[account_id]['launches'][launch_name] = {}

                            deployments[account_id]['launches'][launch_name][region_name] = {
                                'launch_name': launch_name,
                                'portfolio': portfolio.get('DisplayName'),
                                'product': manifest.get('launches', {}).get(launch_name, {}).get('product'),
                                'version': provisioning_artifact_response.get('Name'),
                                'active': provisioning_artifact_response.get('Active'),
                                'region': region_name,
                                'status': status,
                            }
                            output_path = os.path.sep.join([
                                LAUNCHES,
                                account_id,
                                region_name,
                            ])
                            if not os.path.exists(output_path):
                                os.makedirs(output_path)

                            output = os.path.sep.join([output_path, "{}.json".format(provisioned_product.get('Id'))])
                            with open(output, 'w') as f:
                                f.write(json.dumps(
                                    provisioned_product,
                                    indent=4, default=str
                                ))

    table = [
        ['account_id', 'region', 'launch', 'portfolio', 'product', 'expected_version', 'actual_version', 'active',
         'status']
    ]
    for account_id, details in deployment_map.items():
        for launch_name, launch in details.get('launches', {}).items():
            if deployments.get(account_id, {}).get('launches', {}).get(launch_name) is None:
                pass
            else:
                for region, regional_details in deployments[account_id]['launches'][launch_name].items():
                    if regional_details.get('status') == "AVAILABLE":
                        status = Color("{green}" + regional_details.get('status') + "{/green}")
                    else:
                        status = Color("{red}" + regional_details.get('status') + "{/red}")
                    expected_version = launch.get('version')
                    actual_version = regional_details.get('version')
                    if expected_version == actual_version:
                        actual_version = Color("{green}" + actual_version + "{/green}")
                    else:
                        actual_version = Color("{red}" + actual_version + "{/red}")
                    active = regional_details.get('active')
                    if active:
                        active = Color("{green}" + str(active) + "{/green}")
                    else:
                        active = Color("{orange}" + str(active) + "{/orange}")
                    table.append([
                        account_id,
                        region,
                        launch_name,
                        regional_details.get('portfolio'),
                        regional_details.get('product'),
                        expected_version,
                        actual_version,
                        active,
                        status,
                    ])
    click.echo(AsciiTable(table).table)